"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.feeCalculationFunctionsForUBA = exports.getRefundFee = exports.getDepositFee = exports.getEventFee = exports.calculateHistoricalRunningBalance = void 0;
var ethers_1 = require("ethers");
var interfaces_1 = require("../interfaces");
var utils_1 = require("../utils");
var UBAFeeUtility_1 = require("./UBAFeeUtility");
function calculateHistoricalRunningBalance(flows, lastValidatedRunningBalance, lastValidatedIncentiveRunningBalance, lastNetRunningBalanceAdjustment, chainId, tokenSymbol, config) {
    var _a = config.getBalanceTriggerThreshold(chainId, tokenSymbol), upperBoundTriggerHurdle = _a.upperBound, lowerBoundTriggerHurdle = _a.lowerBound;
    var historicalResult = flows.reduce(function (acc, flow) {
        var incentiveFee = flow.incentiveFee;
        if (incentiveFee.mul(-1).gt(acc.incentiveBalance)) {
            throw new Error("Incentive balance will go negative after subtracting flow's incentive reward of ".concat(incentiveFee.toString(), " from incentive balance of ").concat(acc.incentiveBalance.toString()));
        }
        var resultant = {
            netRunningBalanceAdjustment: (0, utils_1.toBN)(acc.netRunningBalanceAdjustment.toString()),
            runningBalance: acc.runningBalance[(0, interfaces_1.isUbaInflow)(flow) ? "add" : "sub"](flow.amount).sub(incentiveFee),
            incentiveBalance: (0, utils_1.max)(ethers_1.ethers.constants.Zero, acc.incentiveBalance.add(incentiveFee)),
        };
        var upperBoundTarget = upperBoundTriggerHurdle.target, upperBoundThreshold = upperBoundTriggerHurdle.threshold;
        var lowerBoundTarget = lowerBoundTriggerHurdle.target, lowerBoundThreshold = lowerBoundTriggerHurdle.threshold;
        if (upperBoundTarget !== undefined &&
            upperBoundThreshold !== undefined &&
            upperBoundThreshold.gt(0) &&
            resultant.runningBalance.gt(upperBoundThreshold)) {
            resultant.netRunningBalanceAdjustment = resultant.netRunningBalanceAdjustment.sub(resultant.runningBalance.sub(upperBoundTarget));
            resultant.runningBalance = upperBoundTarget;
        }
        else if (lowerBoundTarget !== undefined &&
            lowerBoundThreshold !== undefined &&
            lowerBoundThreshold.gt(0) &&
            resultant.runningBalance.lt(lowerBoundThreshold)) {
            resultant.netRunningBalanceAdjustment = resultant.netRunningBalanceAdjustment.add(lowerBoundTarget.sub(resultant.runningBalance));
            resultant.runningBalance = lowerBoundTarget;
        }
        return resultant;
    }, {
        runningBalance: lastValidatedRunningBalance !== null && lastValidatedRunningBalance !== void 0 ? lastValidatedRunningBalance : (0, utils_1.toBN)(0),
        incentiveBalance: lastValidatedIncentiveRunningBalance !== null && lastValidatedIncentiveRunningBalance !== void 0 ? lastValidatedIncentiveRunningBalance : (0, utils_1.toBN)(0),
        netRunningBalanceAdjustment: lastNetRunningBalanceAdjustment !== null && lastNetRunningBalanceAdjustment !== void 0 ? lastNetRunningBalanceAdjustment : (0, utils_1.toBN)(0),
    });
    return historicalResult;
}
exports.calculateHistoricalRunningBalance = calculateHistoricalRunningBalance;
function getEventFee(amount, flowType, lastRunningBalance, lastIncentiveBalance, chainId, config) {
    var flowCurve = config.getBalancingFeeTuples(chainId);
    var balancingFee = (0, UBAFeeUtility_1.computePiecewiseLinearFunction)(flowCurve, lastRunningBalance, lastRunningBalance.add(amount.mul(flowType === "inflow" ? 1 : -1)));
    if (balancingFee.lt(0)) {
        if (lastIncentiveBalance.lte(0)) {
            return {
                balancingFee: ethers_1.ethers.constants.Zero,
            };
        }
        var ubaRewardMultiplier = config.getUbaRewardMultiplier(chainId.toString());
        balancingFee = balancingFee.mul(ubaRewardMultiplier).div(utils_1.fixedPointAdjustment);
        var zeroFeePoint = config.getZeroFeePointOnBalancingFeeCurve(chainId);
        var feesToBringFeePctToZero = (0, UBAFeeUtility_1.computePiecewiseLinearFunction)(flowCurve, zeroFeePoint, lastRunningBalance).abs();
        if (feesToBringFeePctToZero.gt(lastIncentiveBalance)) {
            var discountFactor = lastIncentiveBalance.mul(utils_1.fixedPointAdjustment).div(feesToBringFeePctToZero);
            balancingFee = balancingFee.mul(discountFactor).div(utils_1.fixedPointAdjustment);
        }
    }
    return {
        balancingFee: balancingFee,
    };
}
exports.getEventFee = getEventFee;
function getDepositFee(amount, lastValidatedRunningBalance, lastValidatedIncentiveRunningBalance, chainId, config) {
    return getEventFee(amount, "inflow", lastValidatedRunningBalance, lastValidatedIncentiveRunningBalance, chainId, config);
}
exports.getDepositFee = getDepositFee;
function getRefundFee(amount, lastValidatedRunningBalance, lastValidatedIncentiveRunningBalance, chainId, config) {
    return getEventFee(amount, "outflow", lastValidatedRunningBalance, lastValidatedIncentiveRunningBalance, chainId, config);
}
exports.getRefundFee = getRefundFee;
exports.feeCalculationFunctionsForUBA = {
    deposit: getDepositFee,
    refund: getRefundFee,
};
//# sourceMappingURL=UBAFeeSpokeCalculatorAnalog.js.map