import { __assign, __awaiter, __extends, __generator, __spreadArray } from "tslib";
import { BigNumber, ethers } from "ethers";
import { groupBy } from "lodash";
import { MAX_BIG_INT, assign, isDefined, mapAsync, stringifyJSONWithNumericString, toBN, } from "../utils";
import { paginatedEventQuery, sortEventsAscending, sortEventsAscendingInPlace, spreadEvent, spreadEventWithBlockNumber, } from "../utils/EventUtils";
import { filledSameDeposit, validateFillForDeposit } from "../utils/FlowUtils";
import { ZERO_ADDRESS } from "../constants";
import { getNetworkName } from "../utils/NetworkUtils";
import { getBlockRangeForDepositId, getDepositIdAtBlock } from "../utils/SpokeUtils";
import { BaseAbstractClient } from "./BaseAbstractClient";
/**
 * SpokePoolClient is a client for the SpokePool contract. It is responsible for querying the SpokePool contract
 * for events and storing them in memory. It also provides some convenience methods for querying the stored events.
 */
var SpokePoolClient = /** @class */ (function (_super) {
    __extends(SpokePoolClient, _super);
    /**
     * Creates a new SpokePoolClient.
     * @param logger A logger instance.
     * @param spokePool The SpokePool contract instance that this client will query.
     * @param hubPoolClient An optional HubPoolClient instance. This is used to fetch spoke data that is not stored on the SpokePool contract but is stored on the HubPool contract.
     * @param chainId The chain ID of the chain that this client is querying.
     * @param deploymentBlock The block number that the SpokePool contract was deployed at.
     * @param eventSearchConfig An optional EventSearchConfig object that controls how far back in history the client will search for events. If not provided, the client will only search for events from the deployment block.
     */
    function SpokePoolClient(logger, spokePool, 
    // Can be excluded. This disables some deposit validation.
    hubPoolClient, chainId, deploymentBlock, eventSearchConfig) {
        if (eventSearchConfig === void 0) { eventSearchConfig = { fromBlock: 0, maxBlockLookBack: 0 }; }
        var _this = _super.call(this) || this;
        _this.logger = logger;
        _this.spokePool = spokePool;
        _this.hubPoolClient = hubPoolClient;
        _this.chainId = chainId;
        _this.deploymentBlock = deploymentBlock;
        _this.eventSearchConfig = eventSearchConfig;
        _this.currentTime = 0;
        _this.depositHashes = {};
        _this.depositHashesToFills = {};
        _this.speedUps = {};
        _this.depositRoutes = {};
        _this.tokensBridged = [];
        _this.rootBundleRelays = [];
        _this.relayerRefundExecutions = [];
        _this.earlyDeposits = [];
        _this.queryableEventNames = [];
        _this.earliestDepositIdQueried = Number.MAX_SAFE_INTEGER;
        _this.latestDepositIdQueried = 0;
        _this.firstDepositIdForSpokePool = Number.MAX_SAFE_INTEGER;
        _this.lastDepositIdForSpokePool = Number.MAX_SAFE_INTEGER;
        _this.latestBlockNumber = 0;
        _this.fills = {};
        _this.refundRequests = [];
        _this.firstBlockToSearch = eventSearchConfig.fromBlock;
        _this.latestBlockSearched = eventSearchConfig.fromBlock;
        _this.queryableEventNames = Object.keys(_this._queryableEventNames());
        return _this;
    }
    SpokePoolClient.prototype._queryableEventNames = function () {
        return {
            FundsDeposited: this.spokePool.filters.FundsDeposited(),
            RequestedSpeedUpDeposit: this.spokePool.filters.RequestedSpeedUpDeposit(),
            FilledRelay: this.spokePool.filters.FilledRelay(),
            RefundRequested: this.spokePool.filters.RefundRequested(),
            EnabledDepositRoute: this.spokePool.filters.EnabledDepositRoute(),
            TokensBridged: this.spokePool.filters.TokensBridged(),
            RelayedRootBundle: this.spokePool.filters.RelayedRootBundle(),
            ExecutedRelayerRefundRoot: this.spokePool.filters.ExecutedRelayerRefundRoot(),
        };
    };
    /**
     * Retrieves a list of deposits from the SpokePool contract destined for the given destination chain ID.
     * @param destinationChainId The destination chain ID.
     * @returns A list of deposits.
     */
    SpokePoolClient.prototype.getDepositsForDestinationChain = function (destinationChainId) {
        return Object.values(this.depositHashes).filter(function (deposit) { return deposit.destinationChainId === destinationChainId; });
    };
    /**
     * Retrieves a list of deposits from the SpokePool contract that are associated with this spoke pool.
     * @returns A list of deposits.
     * @note This method returns all deposits, regardless of destination chain ID in sorted order.
     */
    SpokePoolClient.prototype.getDeposits = function () {
        return sortEventsAscendingInPlace(Object.values(this.depositHashes));
    };
    /**
     * Retrieves a list of the tokens that have been bridged.
     * @returns A list of tokens.
     */
    SpokePoolClient.prototype.getTokensBridged = function () {
        return this.tokensBridged;
    };
    /**
     * Retrieves a mapping of tokens and their associated destination chain IDs that can be bridged.
     * @returns A mapping of tokens and their associated destination chain IDs in a nested mapping.
     */
    SpokePoolClient.prototype.getDepositRoutes = function () {
        return this.depositRoutes;
    };
    /**
     * Determines whether a deposit route is enabled for the given origin token and destination chain ID.
     * @param originToken The origin token address.
     * @param destinationChainId The destination chain ID.
     * @returns True if the deposit route is enabled, false otherwise.
     */
    SpokePoolClient.prototype.isDepositRouteEnabled = function (originToken, destinationChainId) {
        var _a, _b;
        return (_b = (_a = this.depositRoutes[originToken]) === null || _a === void 0 ? void 0 : _a[destinationChainId]) !== null && _b !== void 0 ? _b : false;
    };
    /**
     * Retrieves a list of all the available origin tokens that can be bridged.
     * @returns A list of origin tokens.
     */
    SpokePoolClient.prototype.getAllOriginTokens = function () {
        return Object.keys(this.depositRoutes);
    };
    /**
     * Retrieves a list of fills from the SpokePool contract.
     * @returns A list of fills.
     */
    SpokePoolClient.prototype.getFills = function () {
        return sortEventsAscendingInPlace(Object.values(this.fills).flat());
    };
    /**
     * Retrieves a list of fills from a specific origin chain ID.
     * @param originChainId The origin chain ID.
     * @returns A list of fills.
     */
    SpokePoolClient.prototype.getFillsForOriginChain = function (originChainId) {
        return this.fills[originChainId] || [];
    };
    /**
     * Retrieves a list of fills from a specific relayer address.
     * @param relayer The relayer address.
     * @returns A list of fills.
     */
    SpokePoolClient.prototype.getFillsForRelayer = function (relayer) {
        return this.getFills().filter(function (fill) { return fill.relayer === relayer; });
    };
    /**
     * Retrieves a list of fills from a given block range.
     * @param startingBlock The starting block number.
     * @param endingBlock The ending block number.
     * @returns A list of fills.
     */
    SpokePoolClient.prototype.getFillsWithBlockInRange = function (startingBlock, endingBlock) {
        return this.getFills().filter(function (fill) { return fill.blockNumber >= startingBlock && fill.blockNumber <= endingBlock; });
    };
    /**
     * Retrieves a list of refund requests from the SpokePool contract that are within an optional block range.
     * @param fromBlock The starting block number. If not provided, requests will not be filtered by any bounds.
     * @param toBlock The ending block number. If not provided, requests will not be filtered by any bounds.
     * @returns A list of refund requests.
     */
    SpokePoolClient.prototype.getRefundRequests = function (fromBlock, toBlock) {
        return fromBlock === undefined || toBlock === undefined || isNaN(fromBlock) || isNaN(toBlock)
            ? this.refundRequests
            : this.refundRequests.filter(function (request) { return request.blockNumber >= fromBlock && request.blockNumber <= toBlock; });
    };
    /**
     * Retrieves a list of root bundle relays from the SpokePool contract.
     * @returns A list of root bundle relays.
     */
    SpokePoolClient.prototype.getRootBundleRelays = function () {
        return this.rootBundleRelays;
    };
    /**
     * Retrieves the ID of the latest root bundle.
     * @returns The ID of the latest root bundle. This will be 0 if no root bundles have been relayed.
     */
    SpokePoolClient.prototype.getLatestRootBundleId = function () {
        var _a;
        return this.rootBundleRelays.length > 0
            ? ((_a = this.rootBundleRelays[this.rootBundleRelays.length - 1]) === null || _a === void 0 ? void 0 : _a.rootBundleId) + 1
            : 0;
    };
    /**
     * Retrieves a list of relayer refund executions from the SpokePool contract.
     * @returns A list of relayer refund executions.
     */
    SpokePoolClient.prototype.getRelayerRefundExecutions = function () {
        return this.relayerRefundExecutions;
    };
    /**
     * Retrieves a mapping of token addresses to relayer addresses to the amount of refunds that have been executed.
     * @returns A mapping of token addresses to relayer addresses to the amount of refunds that have been executed.
     */
    SpokePoolClient.prototype.getExecutedRefunds = function (relayerRefundRoot) {
        var bundle = this.getRootBundleRelays().find(function (bundle) { return bundle.relayerRefundRoot === relayerRefundRoot; });
        if (bundle === undefined) {
            return {};
        }
        var executedRefundLeaves = this.getRelayerRefundExecutions().filter(function (leaf) { return leaf.rootBundleId === bundle.rootBundleId; });
        var executedRefunds = {};
        for (var _i = 0, executedRefundLeaves_1 = executedRefundLeaves; _i < executedRefundLeaves_1.length; _i++) {
            var refundLeaf = executedRefundLeaves_1[_i];
            var tokenAddress = refundLeaf.l2TokenAddress;
            if (executedRefunds[tokenAddress] === undefined) {
                executedRefunds[tokenAddress] = {};
            }
            var executedTokenRefunds = executedRefunds[tokenAddress];
            for (var i = 0; i < refundLeaf.refundAddresses.length; i++) {
                var relayer = refundLeaf.refundAddresses[i];
                var refundAmount = refundLeaf.refundAmounts[i];
                if (executedTokenRefunds[relayer] === undefined) {
                    executedTokenRefunds[relayer] = ethers.constants.Zero;
                }
                executedTokenRefunds[relayer] = executedTokenRefunds[relayer].add(refundAmount);
            }
        }
        return executedRefunds;
    };
    /**
     * Appends a speed up signature to a specific deposit.
     * @param deposit The deposit to append the speed up signature to.
     * @returns A new deposit instance with the speed up signature appended to the deposit.
     */
    SpokePoolClient.prototype.appendMaxSpeedUpSignatureToDeposit = function (deposit) {
        var _a, _b;
        var maxSpeedUp = (_b = (_a = this.speedUps[deposit.depositor]) === null || _a === void 0 ? void 0 : _a[deposit.depositId]) === null || _b === void 0 ? void 0 : _b.reduce(function (prev, current) {
            return prev.newRelayerFeePct.gt(current.newRelayerFeePct) ? prev : current;
        });
        // We assume that the depositor authorises SpeedUps in isolation of each other, which keeps the relayer
        // logic simple: find the SpeedUp with the highest relayerFeePct, and use all of its fields
        if (!maxSpeedUp || maxSpeedUp.newRelayerFeePct.lte(deposit.relayerFeePct)) {
            return deposit;
        }
        // Return deposit with updated params from the speedup with the highest updated relayer fee pct.
        return __assign(__assign({}, deposit), { speedUpSignature: maxSpeedUp.depositorSignature, newRelayerFeePct: maxSpeedUp.newRelayerFeePct, updatedRecipient: maxSpeedUp.updatedRecipient, updatedMessage: maxSpeedUp.updatedMessage });
    };
    /**
     * Find a corresponding deposit for a given fill.
     * @param fill The fill to find a corresponding deposit for.
     * @returns The corresponding deposit if found, undefined otherwise.
     */
    SpokePoolClient.prototype.getDepositForFill = function (fill, fillFieldsToIgnore) {
        if (fillFieldsToIgnore === void 0) { fillFieldsToIgnore = []; }
        var depositWithMatchingDepositId = this.depositHashes[this.getDepositHash(fill)];
        return validateFillForDeposit(fill, depositWithMatchingDepositId, fillFieldsToIgnore)
            ? depositWithMatchingDepositId
            : undefined;
    };
    /**
     * @dev TODO This function is a bit of a hack for now and its dangerous to leave public because it allows the caller to
     * manipulate internal data that was set at update() time. This is a workaround the current structure where UBAClient
     * is dependent on SpokePoolClient, but one of the SpokePoolClient's internal data structures, `deposits` is dependent
     * on the UBA client state being updated in order to have set correct realizedLpFeePcts. This function is currently
     * designed to be called by the UBA client for each deposit that is loaded and have it reset the realizedLpFeePct
     * equal to the depositBalancingFee plus the LP fee.
     */
    SpokePoolClient.prototype.updateDepositRealizedLpFeePct = function (event, realizedLpFeePct) {
        this.depositHashes[this.getDepositHash(event)] = __assign(__assign({}, this.depositHashes[this.getDepositHash(event)]), { realizedLpFeePct: realizedLpFeePct });
    };
    /**
     * Find the unfilled amount for a given deposit. This is the full deposit amount minus the total filled amount.
     * @param deposit The deposit to find the unfilled amount for.
     * @param fillCount The number of fills that have been applied to this deposit.
     * @param invalidFills The invalid fills that have been applied to this deposit.
     * @returns The unfilled amount.
     */
    SpokePoolClient.prototype.getValidUnfilledAmountForDeposit = function (deposit) {
        var fillsForDeposit = this.depositHashesToFills[this.getDepositHash(deposit)];
        // If no fills then the full amount is remaining.
        if (fillsForDeposit === undefined || fillsForDeposit.length === 0) {
            return { unfilledAmount: toBN(deposit.amount), fillCount: 0, invalidFills: [] };
        }
        var _a = fillsForDeposit.reduce(function (groupedFills, fill) {
            if (validateFillForDeposit(fill, deposit)) {
                groupedFills.validFills.push(fill);
            }
            else {
                groupedFills.invalidFills.push(fill);
            }
            return groupedFills;
        }, { validFills: [], invalidFills: [] }), validFills = _a.validFills, invalidFills = _a.invalidFills;
        // Log any invalid deposits with same deposit id but different params.
        var invalidFillsForDeposit = invalidFills.filter(function (x) { return x.depositId === deposit.depositId; });
        if (invalidFillsForDeposit.length > 0) {
            this.logger.warn({
                at: "SpokePoolClient",
                chainId: this.chainId,
                message: "Invalid fills found matching deposit ID",
                deposit: deposit,
                invalidFills: Object.fromEntries(invalidFillsForDeposit.map(function (x) { return [x.relayer, x]; })),
            });
        }
        // If all fills are invalid we can consider this unfilled.
        if (validFills.length === 0) {
            return { unfilledAmount: toBN(deposit.amount), fillCount: 0, invalidFills: invalidFills };
        }
        // Order fills by totalFilledAmount and then return the first fill's full deposit amount minus total filled amount.
        var fillsOrderedByTotalFilledAmount = validFills.sort(function (fillA, fillB) {
            return fillB.totalFilledAmount.gt(fillA.totalFilledAmount)
                ? 1
                : fillB.totalFilledAmount.lt(fillA.totalFilledAmount)
                    ? -1
                    : 0;
        });
        var lastFill = fillsOrderedByTotalFilledAmount[0];
        return {
            unfilledAmount: toBN(lastFill.amount.sub(lastFill.totalFilledAmount)),
            fillCount: validFills.length,
            invalidFills: invalidFills,
        };
    };
    /**
     * Formulate a hash for a given deposit or fill
     * @param event The deposit or fill to formulate a hash for.
     * @returns The hash.
     * @note This hash is used to match deposits and fills together.
     * @note This hash takes the form of: `${depositId}-${originChainId}`.
     */
    SpokePoolClient.prototype.getDepositHash = function (event) {
        return "".concat(event.depositId, "-").concat(event.originChainId);
    };
    /**
     * Find the block range that contains the deposit ID. This is a binary search that searches for the block range
     * that contains the deposit ID.
     * @param targetDepositId The target deposit ID to search for.
     * @param initLow The initial lower bound of the block range to search.
     * @param initHigh The initial upper bound of the block range to search.
     * @param maxSearches The maximum number of searches to perform. This is used to prevent infinite loops.
     * @returns The block range that contains the deposit ID.
     * @note  // We want to find the block range that satisfies these conditions:
     *        // - the low block has deposit count <= targetDepositId
     *        // - the high block has a deposit count > targetDepositId.
     *        // This way the caller can search for a FundsDeposited event between [low, high] that will always
     *        // contain the event emitted when deposit ID was incremented to targetDepositId + 1. This is the same transaction
     *        // where the deposit with deposit ID = targetDepositId was created.
     */
    SpokePoolClient.prototype._getBlockRangeForDepositId = function (targetDepositId, initLow, initHigh, maxSearches) {
        return getBlockRangeForDepositId(targetDepositId, initLow, initHigh, maxSearches, this);
    };
    /**
     * Finds the deposit id at a specific block number.
     * @param blockTag The block number to search for the deposit ID at.
     * @returns The deposit ID.
     */
    SpokePoolClient.prototype._getDepositIdAtBlock = function (blockTag) {
        return getDepositIdAtBlock(this.spokePool, blockTag);
    };
    /**
     * Queries the SpokePool contract for a list of historical fills that match the given fill and deposit.
     * @param fill The fill to match.
     * @param deposit The deposit to match.
     * @param toBlock The block number to search up to.
     * @returns A list of fills that match the given fill and deposit.
     */
    SpokePoolClient.prototype.queryHistoricalMatchingFills = function (fill, deposit, toBlock) {
        return __awaiter(this, void 0, void 0, function () {
            var searchConfig;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        searchConfig = {
                            fromBlock: this.deploymentBlock,
                            toBlock: toBlock,
                            maxBlockLookBack: this.eventSearchConfig.maxBlockLookBack,
                        };
                        return [4 /*yield*/, this.queryFillsInBlockRange(fill, searchConfig)];
                    case 1: return [2 /*return*/, (_a.sent()).filter(function (_fill) {
                            return validateFillForDeposit(_fill, deposit);
                        })];
                }
            });
        });
    };
    /**
     * Queries the SpokePool contract for a list of fills that match the given fill.
     * @param fill The fill to match.
     * @param searchConfig The search configuration.
     * @returns A Promise that resolves to a list of fills that match the given fill.
     */
    SpokePoolClient.prototype.queryFillsInBlockRange = function (matchingFill, searchConfig) {
        return __awaiter(this, void 0, void 0, function () {
            var query, fills;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, paginatedEventQuery(this.spokePool, this.spokePool.filters.FilledRelay(undefined, // amount
                        undefined, // totalFilledAmount
                        undefined, // fillAmount
                        undefined, // repaymentChainId
                        matchingFill.originChainId, // originChainId
                        undefined, // destinationChainId
                        undefined, // relayerFeePct
                        undefined, // realizedLpFeePct
                        matchingFill.depositId, // depositId
                        undefined, // destinationToken
                        undefined, // relayer
                        matchingFill.depositor, // depositor
                        undefined, // recipient
                        undefined, // message
                        undefined // updatableRelayData
                        ), searchConfig)];
                    case 1:
                        query = _a.sent();
                        fills = query.map(function (event) { return spreadEventWithBlockNumber(event); });
                        return [2 /*return*/, sortEventsAscending(fills.filter(function (_fill) { return filledSameDeposit(_fill, matchingFill); }))];
                }
            });
        });
    };
    /**
     * Performs an update to refresh the state of this client. This will query the SpokePool contract for new events
     * and store them in memory. This method is the primary method for updating the state of this client.
     * @param eventsToQuery An optional list of events to query. If not provided, all events will be queried.
     * @returns A Promise that resolves to a SpokePoolUpdate object.
     */
    SpokePoolClient.prototype._update = function (eventsToQuery) {
        var _a;
        return __awaiter(this, void 0, void 0, function () {
            var firstDepositId, latestBlockNumber, searchConfig, eventSearchConfigs, timerStart, _b, numberOfDeposits, currentTime, events, errMsg, blocks, isUBAActivated, blockNumbers, _c, _d;
            var _this = this;
            return __generator(this, function (_e) {
                switch (_e.label) {
                    case 0:
                        firstDepositId = this.firstDepositIdForSpokePool;
                        if (!(firstDepositId === Number.MAX_SAFE_INTEGER)) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.spokePool.numberOfDeposits({ blockTag: this.deploymentBlock })];
                    case 1:
                        firstDepositId = _e.sent();
                        if (isNaN(firstDepositId) || firstDepositId < 0) {
                            throw new Error("SpokePoolClient::update: Invalid first deposit id (".concat(firstDepositId, ")"));
                        }
                        _e.label = 2;
                    case 2: return [4 /*yield*/, this.spokePool.provider.getBlockNumber()];
                    case 3:
                        latestBlockNumber = _e.sent();
                        if (isNaN(latestBlockNumber) || latestBlockNumber < this.latestBlockNumber) {
                            throw new Error("SpokePoolClient::update: latestBlockNumber ".concat(latestBlockNumber, " < ").concat(this.latestBlockNumber));
                        }
                        searchConfig = {
                            fromBlock: this.firstBlockToSearch,
                            toBlock: this.eventSearchConfig.toBlock || latestBlockNumber,
                            maxBlockLookBack: this.eventSearchConfig.maxBlockLookBack,
                        };
                        if (searchConfig.fromBlock > searchConfig.toBlock) {
                            this.log("warn", "Invalid update() searchConfig.", { searchConfig: searchConfig });
                            return [2 /*return*/, { success: false }];
                        }
                        eventSearchConfigs = eventsToQuery.map(function (eventName) {
                            if (!_this.queryableEventNames.includes(eventName)) {
                                throw new Error("SpokePoolClient: Cannot query unrecognised SpokePool event name: ".concat(eventName));
                            }
                            var _searchConfig = __assign({}, searchConfig); // shallow copy
                            // By default, an event's query range is controlled by the `eventSearchConfig` passed in during instantiation.
                            // However, certain events have special overriding requirements to their search ranges:
                            // - EnabledDepositRoute: The full history is always required, so override the requested fromBlock.
                            if (eventName === "EnabledDepositRoute" && !_this.isUpdated) {
                                _searchConfig.fromBlock = _this.deploymentBlock;
                            }
                            return {
                                filter: _this._queryableEventNames()[eventName],
                                searchConfig: _searchConfig,
                            };
                        });
                        this.log("debug", "Updating SpokePool client for chain ".concat(this.chainId), {
                            eventsToQuery: eventsToQuery,
                            searchConfig: searchConfig,
                            spokePool: this.spokePool.address,
                        });
                        timerStart = Date.now();
                        return [4 /*yield*/, Promise.all(__spreadArray([
                                this.spokePool.numberOfDeposits({ blockTag: searchConfig.toBlock }),
                                this.spokePool.getCurrentTime({ blockTag: searchConfig.toBlock })
                            ], eventSearchConfigs.map(function (config) { return paginatedEventQuery(_this.spokePool, config.filter, config.searchConfig); }), true))];
                    case 4:
                        _b = _e.sent(), numberOfDeposits = _b[0], currentTime = _b[1], events = _b.slice(2);
                        this.log("debug", "Time to query new events from RPC for ".concat(this.chainId, ": ").concat(Date.now() - timerStart, " ms"));
                        if (!BigNumber.isBigNumber(currentTime) || currentTime.lt(this.currentTime)) {
                            errMsg = BigNumber.isBigNumber(currentTime)
                                ? "currentTime: ".concat(currentTime, " < ").concat(toBN(this.currentTime))
                                : "currentTime is not a BigNumber: ".concat(JSON.stringify(currentTime));
                            throw new Error("SpokePoolClient::update: ".concat(errMsg));
                        }
                        // Sort all events to ensure they are stored in a consistent order.
                        events.forEach(function (events) { return sortEventsAscendingInPlace(events); });
                        blocks = undefined;
                        isUBAActivated = isDefined((_a = this.hubPoolClient) === null || _a === void 0 ? void 0 : _a.configStoreClient.getUBAActivationBlock());
                        if (!isUBAActivated) return [3 /*break*/, 6];
                        blockNumbers = Array.from(new Set(["FundsDeposited", "FilledRelay", "RefundRequested"]
                            .filter(function (eventName) { return eventsToQuery.includes(eventName); })
                            .map(function (eventName) {
                            var idx = eventsToQuery.indexOf(eventName);
                            // tsc needs type hints on this map...
                            return events[idx].map(function (_a) {
                                var blockNumber = _a.blockNumber;
                                return blockNumber;
                            });
                        })
                            .flat()));
                        _d = (_c = Object).fromEntries;
                        return [4 /*yield*/, mapAsync(blockNumbers, function (blockNumber) { return __awaiter(_this, void 0, void 0, function () {
                                var block;
                                return __generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0: return [4 /*yield*/, this.spokePool.provider.getBlock(blockNumber)];
                                        case 1:
                                            block = _a.sent();
                                            return [2 /*return*/, [blockNumber, block]];
                                    }
                                });
                            }); })];
                    case 5:
                        blocks = _d.apply(_c, [_e.sent()]);
                        _e.label = 6;
                    case 6: return [2 /*return*/, {
                            success: true,
                            currentTime: currentTime.toNumber(),
                            firstDepositId: firstDepositId,
                            latestBlockNumber: latestBlockNumber,
                            latestDepositId: Math.max(numberOfDeposits - 1, 0),
                            searchEndBlock: searchConfig.toBlock,
                            events: events,
                            blocks: blocks,
                        }];
                }
            });
        });
    };
    SpokePoolClient.prototype._isEarlyDeposit = function (depositEvent, currentTime) {
        return depositEvent.args.quoteTimestamp > currentTime;
    };
    /**
     * A wrapper over the `_update` method that handles errors and logs. This method additionally calls into the
     * HubPoolClient to update the state of this client with data from the HubPool contract.
     * @param eventsToQuery An optional list of events to query. If not provided, all events will be queried.
     * @returns A Promise that resolves to a SpokePoolUpdate object.
     * @note This method is the primary method for updating the state of this client externally.
     * @see _update
     */
    SpokePoolClient.prototype.update = function (eventsToQuery) {
        var _a, _b;
        if (eventsToQuery === void 0) { eventsToQuery = this.queryableEventNames; }
        return __awaiter(this, void 0, void 0, function () {
            var update, queryResults, blocks, currentTime, _i, _c, event_1, allDeposits, _d, _e, earlyDeposits, _f, depositEvents, dataForQuoteTime, _g, _h, _j, index, event_2, rawDeposit, deposit, _k, _l, speedUpEvents, _m, speedUpEvents_1, event_3, speedUp, depositHash, depositDataAssociatedWithSpeedUp, fillEvents, _o, fillEvents_1, event_4, rawFill, fill, refundRequests, _p, refundRequests_1, event_5, rawRefundRequest, refundRequest, enableDepositsEvents, _q, enableDepositsEvents_1, event_6, enableDeposit, relayedRootBundleEvents, _r, relayedRootBundleEvents_1, event_7, executedRelayerRefundRootEvents, _s, executedRelayerRefundRootEvents_1, event_8, executedRefund;
            var _this = this;
            return __generator(this, function (_t) {
                switch (_t.label) {
                    case 0:
                        if (this.hubPoolClient !== null && !this.hubPoolClient.isUpdated) {
                            throw new Error("HubPoolClient not updated");
                        }
                        return [4 /*yield*/, this._update(eventsToQuery)];
                    case 1:
                        update = _t.sent();
                        if (!update.success) {
                            // This failure only occurs if the RPC searchConfig is miscomputed, and has only been seen in the hardhat test
                            // environment. Normal failures will throw instead. This is therefore an unfortunate workaround until we can
                            // understand why we see this in test. @todo: Resolve.
                            return [2 /*return*/];
                        }
                        queryResults = update.events, blocks = update.blocks, currentTime = update.currentTime;
                        if (eventsToQuery.includes("TokensBridged")) {
                            for (_i = 0, _c = queryResults[eventsToQuery.indexOf("TokensBridged")]; _i < _c.length; _i++) {
                                event_1 = _c[_i];
                                this.tokensBridged.push(spreadEventWithBlockNumber(event_1));
                            }
                        }
                        if (!eventsToQuery.includes("FundsDeposited")) return [3 /*break*/, 9];
                        allDeposits = __spreadArray(__spreadArray([], queryResults[eventsToQuery.indexOf("FundsDeposited")], true), this.earlyDeposits, true);
                        _d = groupBy(allDeposits, function (depositEvent) {
                            if (_this._isEarlyDeposit(depositEvent, currentTime)) {
                                var args = depositEvent.args, transactionHash = depositEvent.transactionHash;
                                _this.logger.debug({
                                    at: "SpokePoolClient#update",
                                    message: "Deferring early deposit event.",
                                    currentTime: currentTime,
                                    deposit: { args: args, transactionHash: transactionHash },
                                });
                                return "earlyDeposits";
                            }
                            else {
                                return "depositEvents";
                            }
                        }), _e = _d.earlyDeposits, earlyDeposits = _e === void 0 ? [] : _e, _f = _d.depositEvents, depositEvents = _f === void 0 ? [] : _f;
                        this.earlyDeposits = earlyDeposits;
                        return [4 /*yield*/, Promise.all(depositEvents.map(function (event) { return _this.computeRealizedLpFeePct(event); }))];
                    case 2:
                        dataForQuoteTime = _t.sent();
                        // Now add any newly fetched events from RPC.
                        if (depositEvents.length > 0) {
                            this.log("debug", "Using ".concat(depositEvents.length, " newly queried deposit events for chain ").concat(this.chainId), {
                                earliestEvent: depositEvents[0].blockNumber,
                            });
                        }
                        _g = 0, _h = Array.from(depositEvents.entries());
                        _t.label = 3;
                    case 3:
                        if (!(_g < _h.length)) return [3 /*break*/, 9];
                        _j = _h[_g], index = _j[0], event_2 = _j[1];
                        rawDeposit = spreadEventWithBlockNumber(event_2);
                        deposit = __assign(__assign({}, rawDeposit), { realizedLpFeePct: dataForQuoteTime[index].realizedLpFeePct, destinationToken: this.getDestinationTokenForDeposit(rawDeposit), quoteBlockNumber: dataForQuoteTime[index].quoteBlock, blockTimestamp: 0 });
                        if (!isDefined(blocks)) return [3 /*break*/, 7];
                        _k = deposit;
                        if (!((_b = (_a = blocks[event_2.blockNumber]) === null || _a === void 0 ? void 0 : _a.timestamp) !== null && _b !== void 0)) return [3 /*break*/, 4];
                        _l = _b;
                        return [3 /*break*/, 6];
                    case 4: return [4 /*yield*/, event_2.getBlock()];
                    case 5:
                        _l = (_t.sent()).timestamp;
                        _t.label = 6;
                    case 6:
                        _k.blockTimestamp = _l;
                        _t.label = 7;
                    case 7:
                        assign(this.depositHashes, [this.getDepositHash(deposit)], deposit);
                        if (deposit.depositId < this.earliestDepositIdQueried) {
                            this.earliestDepositIdQueried = deposit.depositId;
                        }
                        if (deposit.depositId > this.latestDepositIdQueried) {
                            this.latestDepositIdQueried = deposit.depositId;
                        }
                        _t.label = 8;
                    case 8:
                        _g++;
                        return [3 /*break*/, 3];
                    case 9:
                        // TODO: When validating fills with deposits for the purposes of UBA flows, do we need to consider
                        // speed ups as well? For example, do we need to also consider that the speed up is before the fill
                        // timestamp to be applied for the fill? My brain hurts.
                        // Update deposits with speed up requests from depositor.
                        if (eventsToQuery.includes("RequestedSpeedUpDeposit")) {
                            speedUpEvents = queryResults[eventsToQuery.indexOf("RequestedSpeedUpDeposit")];
                            for (_m = 0, speedUpEvents_1 = speedUpEvents; _m < speedUpEvents_1.length; _m++) {
                                event_3 = speedUpEvents_1[_m];
                                speedUp = __assign(__assign({}, spreadEvent(event_3.args)), { originChainId: this.chainId });
                                assign(this.speedUps, [speedUp.depositor, speedUp.depositId], [speedUp]);
                                depositHash = this.getDepositHash(speedUp);
                                depositDataAssociatedWithSpeedUp = this.depositHashes[depositHash];
                                if (isDefined(depositDataAssociatedWithSpeedUp)) {
                                    this.depositHashes[depositHash] = this.appendMaxSpeedUpSignatureToDeposit(depositDataAssociatedWithSpeedUp);
                                }
                            }
                        }
                        if (eventsToQuery.includes("FilledRelay")) {
                            fillEvents = queryResults[eventsToQuery.indexOf("FilledRelay")];
                            if (fillEvents.length > 0) {
                                this.log("debug", "Using ".concat(fillEvents.length, " newly queried fill events for chain ").concat(this.chainId), {
                                    earliestEvent: fillEvents[0].blockNumber,
                                });
                            }
                            for (_o = 0, fillEvents_1 = fillEvents; _o < fillEvents_1.length; _o++) {
                                event_4 = fillEvents_1[_o];
                                rawFill = spreadEventWithBlockNumber(event_4);
                                fill = __assign(__assign({}, rawFill), { blockTimestamp: 0 });
                                // Override the default blockTimestamp of 0 only if the UBA is active and we have pre-queried block times
                                // for each event.
                                if (isDefined(blocks)) {
                                    fill.blockTimestamp = blocks[event_4.blockNumber].timestamp;
                                }
                                assign(this.fills, [fill.originChainId], [fill]);
                                assign(this.depositHashesToFills, [this.getDepositHash(fill)], [fill]);
                            }
                        }
                        // @note: In Across 2.5, callers will simultaneously request [FundsDeposited, FilledRelay, RefundsRequested].
                        // The list of events is always pre-sorted, so rather than splitting them out individually, it might make sense to
                        // evaluate them as a single group, to avoid having to re-merge and sequence again afterwards.
                        if (eventsToQuery.includes("RefundRequested")) {
                            refundRequests = queryResults[eventsToQuery.indexOf("RefundRequested")];
                            if (refundRequests.length > 0) {
                                this.log("debug", "Found ".concat(refundRequests.length, " new relayer refund requests on chain ").concat(this.chainId), {
                                    earliestEvent: refundRequests[0].blockNumber,
                                });
                            }
                            for (_p = 0, refundRequests_1 = refundRequests; _p < refundRequests_1.length; _p++) {
                                event_5 = refundRequests_1[_p];
                                rawRefundRequest = spreadEventWithBlockNumber(event_5);
                                refundRequest = __assign(__assign({}, rawRefundRequest), { repaymentChainId: this.chainId, blockTimestamp: 0 });
                                // Override the default blockTimestamp of 0 only if the UBA is active and we have pre-queried block times
                                // for each event.
                                if (isDefined(blocks)) {
                                    refundRequest.blockTimestamp = blocks[event_5.blockNumber].timestamp;
                                }
                                this.refundRequests.push(refundRequest);
                            }
                        }
                        if (eventsToQuery.includes("EnabledDepositRoute")) {
                            enableDepositsEvents = queryResults[eventsToQuery.indexOf("EnabledDepositRoute")];
                            for (_q = 0, enableDepositsEvents_1 = enableDepositsEvents; _q < enableDepositsEvents_1.length; _q++) {
                                event_6 = enableDepositsEvents_1[_q];
                                enableDeposit = spreadEvent(event_6.args);
                                assign(this.depositRoutes, [enableDeposit.originToken, enableDeposit.destinationChainId], enableDeposit.enabled);
                            }
                        }
                        if (eventsToQuery.includes("RelayedRootBundle")) {
                            relayedRootBundleEvents = queryResults[eventsToQuery.indexOf("RelayedRootBundle")];
                            for (_r = 0, relayedRootBundleEvents_1 = relayedRootBundleEvents; _r < relayedRootBundleEvents_1.length; _r++) {
                                event_7 = relayedRootBundleEvents_1[_r];
                                this.rootBundleRelays.push(spreadEventWithBlockNumber(event_7));
                            }
                        }
                        if (eventsToQuery.includes("ExecutedRelayerRefundRoot")) {
                            executedRelayerRefundRootEvents = queryResults[eventsToQuery.indexOf("ExecutedRelayerRefundRoot")];
                            for (_s = 0, executedRelayerRefundRootEvents_1 = executedRelayerRefundRootEvents; _s < executedRelayerRefundRootEvents_1.length; _s++) {
                                event_8 = executedRelayerRefundRootEvents_1[_s];
                                executedRefund = spreadEventWithBlockNumber(event_8);
                                executedRefund.l2TokenAddress = SpokePoolClient.getExecutedRefundLeafL2Token(executedRefund.chainId, executedRefund.l2TokenAddress);
                                this.relayerRefundExecutions.push(executedRefund);
                            }
                        }
                        // Next iteration should start off from where this one ended.
                        this.currentTime = currentTime;
                        this.firstDepositIdForSpokePool = update.firstDepositId;
                        this.latestBlockNumber = update.latestBlockNumber;
                        this.lastDepositIdForSpokePool = update.latestDepositId;
                        this.firstBlockToSearch = update.searchEndBlock + 1;
                        this.latestBlockSearched = update.searchEndBlock;
                        this.isUpdated = true;
                        this.log("debug", "SpokePool client for chain ".concat(this.chainId, " updated!"), {
                            nextFirstBlockToSearch: this.firstBlockToSearch,
                        });
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Retrieves the l2TokenAddress for a given executed refund leaf.
     * @param chainId The chain ID of the executed refund leaf.
     * @param eventL2Token The l2TokenAddress of the executed refund leaf.
     * @returns The l2TokenAddress of the executed refund leaf.
     */
    SpokePoolClient.getExecutedRefundLeafL2Token = function (chainId, eventL2Token) {
        // If execution of WETH refund leaf occurred on an OVM spoke pool, then we'll convert its l2Token from the native
        // token address to the wrapped token address. This is because the OVM_SpokePool modifies the l2TokenAddress prop
        // in _bridgeTokensToHubPool before emitting the ExecutedRelayerRefundLeaf event.
        // Here is the contract code referenced:
        // - https://github.com/across-protocol/contracts-v2/blob/954528a4620863d1c868e54a370fd8556d5ed05c/contracts/Ovm_SpokePool.sol#L142
        if ((chainId === 10 || chainId === 8453) &&
            eventL2Token.toLowerCase() === "0xdeaddeaddeaddeaddeaddeaddeaddeaddead0000") {
            return "0x4200000000000000000000000000000000000006";
        }
        else if (chainId === 288 && eventL2Token.toLowerCase() === "0x4200000000000000000000000000000000000006") {
            return "0xDeadDeAddeAddEAddeadDEaDDEAdDeaDDeAD0000";
        }
        else {
            return eventL2Token;
        }
    };
    /**
     * Computes the realized LP fee percentage for a given deposit.
     * @param depositEvent The deposit event to compute the realized LP fee percentage for.
     * @returns The realized LP fee percentage.
     */
    SpokePoolClient.prototype.computeRealizedLpFeePct = function (depositEvent) {
        // If no hub pool client, we're using this for testing. So set quote block very high
        // so that if its ever used to look up a configuration for a block, it will always match with some
        // configuration because the quote block will always be greater than the updated config event block height.
        if (this.hubPoolClient === null) {
            return { realizedLpFeePct: toBN(0), quoteBlock: MAX_BIG_INT.toNumber() };
        }
        var deposit = {
            amount: depositEvent.args.amount,
            originChainId: Number(depositEvent.args.originChainId),
            destinationChainId: Number(depositEvent.args.destinationChainId),
            originToken: depositEvent.args.originToken,
            quoteTimestamp: depositEvent.args.quoteTimestamp,
            blockNumber: depositEvent.blockNumber,
        };
        var l1Token = this.hubPoolClient.getL1TokenForDeposit(deposit);
        return this.hubPoolClient.computeRealizedLpFeePct(deposit, l1Token);
    };
    /**
     * Retrieves the destination token for a given deposit.
     * @param deposit The deposit to retrieve the destination token for.
     * @returns The destination token.
     */
    SpokePoolClient.prototype.getDestinationTokenForDeposit = function (deposit) {
        // If there is no rate model client return address(0).
        if (!this.hubPoolClient) {
            return ZERO_ADDRESS;
        }
        return this.hubPoolClient.getDestinationTokenForDeposit(deposit);
    };
    /**
     * Performs a log for a specific level, message and data.
     * @param level The log level.
     * @param message The log message.
     * @param data Optional data to log.
     */
    SpokePoolClient.prototype.log = function (level, message, data) {
        this.logger[level](__assign({ at: "SpokePoolClient", chainId: this.chainId, message: message }, data));
    };
    /**
     * Retrieves the current time from the SpokePool contract.
     * @returns The current time.
     */
    SpokePoolClient.prototype.getCurrentTime = function () {
        return this.currentTime;
    };
    /**
     * Finds a deposit for a given deposit ID, destination chain ID and depositor address. This method will search for
     * the deposit in the SpokePool contract and return it if found. If the deposit is not found, this method will
     * perform a binary search to find the block range that contains the deposit ID and then perform an eth_getLogs
     * call to find the deposit.
     * @param depositId The deposit ID to find.
     * @param destinationChainId The destination chain ID to find.
     * @param depositor The depositor address to find.
     * @returns The deposit if found.
     * @note This method is used to find deposits that are outside of the search range of this client.
     */
    SpokePoolClient.prototype.findDeposit = function (depositId, destinationChainId, depositor) {
        return __awaiter(this, void 0, void 0, function () {
            var searchBounds, tStart, query, tStop, event, srcChain, dstChain, partialDeposit, _a, realizedLpFeePct, quoteBlockNumber, deposit;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4 /*yield*/, this._getBlockRangeForDepositId(depositId, this.deploymentBlock, this.latestBlockNumber, 7)];
                    case 1:
                        searchBounds = _b.sent();
                        tStart = Date.now();
                        return [4 /*yield*/, paginatedEventQuery(this.spokePool, this.spokePool.filters.FundsDeposited(null, null, destinationChainId, null, depositId, null, null, null, depositor, null), {
                                fromBlock: searchBounds.low,
                                toBlock: searchBounds.high,
                                maxBlockLookBack: this.eventSearchConfig.maxBlockLookBack,
                            })];
                    case 2:
                        query = _b.sent();
                        tStop = Date.now();
                        event = query.find(function (deposit) { return deposit.args.depositId === depositId; });
                        if (event === undefined) {
                            srcChain = getNetworkName(this.chainId);
                            dstChain = getNetworkName(destinationChainId);
                            throw new Error("Could not find deposit ".concat(depositId, " for ").concat(dstChain, " fill") +
                                " between ".concat(srcChain, " blocks [").concat(searchBounds.low, ", ").concat(searchBounds.high, "]"));
                        }
                        partialDeposit = spreadEventWithBlockNumber(event);
                        return [4 /*yield*/, this.computeRealizedLpFeePct(event)];
                    case 3:
                        _a = _b.sent(), realizedLpFeePct = _a.realizedLpFeePct, quoteBlockNumber = _a.quoteBlock;
                        deposit = __assign(__assign({}, partialDeposit), { realizedLpFeePct: realizedLpFeePct, destinationToken: this.getDestinationTokenForDeposit(partialDeposit), quoteBlockNumber: quoteBlockNumber });
                        this.logger.debug({
                            at: "SpokePoolClient#findDeposit",
                            message: "Located deposit outside of SpokePoolClient's search range",
                            deposit: deposit,
                            elapsedMs: tStop - tStart,
                        });
                        return [2 /*return*/, deposit];
                }
            });
        });
    };
    SpokePoolClient.prototype.updateFromJSON = function (spokePoolClientState) {
        var keysToUpdate = Object.keys(spokePoolClientState);
        this.logger.debug({
            at: "SpokePoolClient",
            message: "Updating SpokePool client from JSON",
            keys: keysToUpdate,
        });
        if (keysToUpdate.length === 0) {
            return;
        }
        this.currentTime = spokePoolClientState.currentTime || this.currentTime;
        this.depositHashes = Object.entries(spokePoolClientState.depositHashes || {}).reduce(function (acc, _a) {
            var _b;
            var hash = _a[0], deposit = _a[1];
            return (__assign(__assign({}, acc), (_b = {}, _b[hash] = __assign(__assign({}, deposit), { amount: BigNumber.from(deposit.amount), relayerFeePct: BigNumber.from(deposit.relayerFeePct), realizedLpFeePct: deposit.relayerFeePct ? BigNumber.from(deposit.realizedLpFeePct) : undefined, newRelayerFeePct: deposit.newRelayerFeePct ? BigNumber.from(deposit.newRelayerFeePct) : undefined }), _b)));
        }, {});
        this.depositHashesToFills = Object.entries(spokePoolClientState.depositHashesToFills || {}).reduce(function (acc, _a) {
            var _b;
            var hash = _a[0], fills = _a[1];
            return (__assign(__assign({}, acc), (_b = {}, _b[hash] = fills.map(function (fill) { return (__assign(__assign({}, fill), { amount: BigNumber.from(fill.amount), totalFilledAmount: BigNumber.from(fill.totalFilledAmount), fillAmount: BigNumber.from(fill.fillAmount), relayerFeePct: BigNumber.from(fill.relayerFeePct), realizedLpFeePct: BigNumber.from(fill.realizedLpFeePct), updatableRelayData: __assign(__assign({}, fill.updatableRelayData), { relayerFeePct: BigNumber.from(fill.updatableRelayData.relayerFeePct), payoutAdjustmentPct: BigNumber.from(fill.updatableRelayData.payoutAdjustmentPct) }) })); }), _b)));
        }, {});
        this.speedUps = Object.entries(spokePoolClientState.speedUps || {}).reduce(function (acc, _a) {
            var _b;
            var depositor = _a[0], speedUpsMap = _a[1];
            var speedUps = Object.entries(speedUpsMap).reduce(function (acc, _a) {
                var _b;
                var depositId = _a[0], speedUps = _a[1];
                return (__assign(__assign({}, acc), (_b = {}, _b[depositId] = speedUps.map(function (speedUp) { return (__assign(__assign({}, speedUp), { newRelayerFeePct: BigNumber.from(speedUp.newRelayerFeePct) })); }), _b)));
            }, {});
            return __assign(__assign({}, acc), (_b = {}, _b[depositor] = speedUps, _b));
        }, {});
        this.depositRoutes = spokePoolClientState.depositRoutes || {};
        this.tokensBridged = (spokePoolClientState.tokensBridged || []).map(function (tokenBridged) { return (__assign(__assign({}, tokenBridged), { amountToReturn: BigNumber.from(tokenBridged.amountToReturn) })); });
        this.rootBundleRelays = spokePoolClientState.rootBundleRelays || [];
        this.relayerRefundExecutions = (spokePoolClientState.relayerRefundExecutions || []).map(function (refundExecution) { return (__assign(__assign({}, refundExecution), { amountToReturn: BigNumber.from(refundExecution.amountToReturn), refundAmounts: refundExecution.refundAmounts.map(function (refundAmount) { return BigNumber.from(refundAmount); }) })); });
        this.earlyDeposits = (spokePoolClientState.earlyDeposits || []).map(function (deposit) { return (__assign(__assign({}, deposit), { amount: BigNumber.from(deposit.amount), originChainId: BigNumber.from(deposit.originChainId), destinationChainId: BigNumber.from(deposit.destinationChainId), relayerFeePct: BigNumber.from(deposit.relayerFeePct) })); });
        this.queryableEventNames = spokePoolClientState.queryableEventNames || this.queryableEventNames;
        this.earliestDepositIdQueried = spokePoolClientState.earliestDepositIdQueried || this.earliestDepositIdQueried;
        this.latestDepositIdQueried = spokePoolClientState.latestDepositIdQueried || this.latestDepositIdQueried;
        this.firstBlockToSearch = spokePoolClientState.firstBlockToSearch || this.firstBlockToSearch;
        this.latestBlockSearched = spokePoolClientState.latestBlockSearched || this.latestBlockSearched;
        this.fills = Object.entries(spokePoolClientState.fills || []).reduce(function (acc, _a) {
            var _b;
            var chainId = _a[0], fills = _a[1];
            return (__assign(__assign({}, acc), (_b = {}, _b[chainId] = fills.map(function (fill) { return (__assign(__assign({}, fill), { amount: BigNumber.from(fill.amount), totalFilledAmount: BigNumber.from(fill.totalFilledAmount), fillAmount: BigNumber.from(fill.fillAmount), relayerFeePct: BigNumber.from(fill.relayerFeePct), realizedLpFeePct: BigNumber.from(fill.realizedLpFeePct), updatableRelayData: __assign(__assign({}, fill.updatableRelayData), { relayerFeePct: BigNumber.from(fill.updatableRelayData.relayerFeePct), payoutAdjustmentPct: BigNumber.from(fill.updatableRelayData.payoutAdjustmentPct) }) })); }), _b)));
        }, {});
        this.refundRequests = (spokePoolClientState.refundRequests || []).map(function (refundRequest) { return (__assign(__assign({}, refundRequest), { amount: BigNumber.from(refundRequest.amount), realizedLpFeePct: BigNumber.from(refundRequest.realizedLpFeePct), previousIdenticalRequests: BigNumber.from(refundRequest.previousIdenticalRequests), fillBlock: BigNumber.from(refundRequest.fillBlock) })); });
        this.isUpdated = true;
    };
    SpokePoolClient.prototype.toJSON = function () {
        return {
            chainId: this.chainId,
            deploymentBlock: this.deploymentBlock,
            eventSearchConfig: this.eventSearchConfig,
            currentTime: this.currentTime,
            depositHashes: JSON.parse(stringifyJSONWithNumericString(this.depositHashes)),
            depositHashesToFills: JSON.parse(stringifyJSONWithNumericString(this.depositHashesToFills)),
            speedUps: JSON.parse(stringifyJSONWithNumericString(this.speedUps)),
            depositRoutes: this.depositRoutes,
            tokensBridged: JSON.parse(stringifyJSONWithNumericString(this.tokensBridged)),
            rootBundleRelays: this.rootBundleRelays,
            relayerRefundExecutions: JSON.parse(stringifyJSONWithNumericString(this.relayerRefundExecutions)),
            earlyDeposits: JSON.parse(stringifyJSONWithNumericString(this.earlyDeposits)),
            queryableEventNames: this.queryableEventNames,
            earliestDepositIdQueried: this.earliestDepositIdQueried,
            earliestDepositIdForSpokePool: this.firstDepositIdForSpokePool,
            latestDepositIdQueried: this.latestDepositIdQueried,
            latestDepositIdForSpokePool: this.lastDepositIdForSpokePool,
            firstBlockToSearch: this.firstBlockToSearch,
            latestBlockSearched: this.latestBlockSearched,
            isUpdated: this.isUpdated,
            fills: JSON.parse(stringifyJSONWithNumericString(this.fills)),
            refundRequests: JSON.parse(stringifyJSONWithNumericString(this.refundRequests)),
        };
    };
    return SpokePoolClient;
}(BaseAbstractClient));
export { SpokePoolClient };
//# sourceMappingURL=SpokePoolClient.js.map